@@ -1,9 +1,9 @@
-/* Copyright 2001-2005 The Apache Software Foundation or its licensors, as
- * applicable.
- *
- * Licensed under the Apache License, Version 2.0 (the "License");
- * you may not use this file except in compliance with the License.
- * You may obtain a copy of the License at
+/* Licensed to the Apache Software Foundation (ASF) under one or more
+ * contributor license agreements.  See the NOTICE file distributed with
+ * this work for additional information regarding copyright ownership.
+ * The ASF licenses this file to You under the Apache License, Version 2.0
+ * (the "License"); you may not use this file except in compliance with
+ * the License.  You may obtain a copy of the License at
  *
  *     http://www.apache.org/licenses/LICENSE-2.0
  *
@@ -14,19 +14,36 @@
  * limitations under the License.
  */
 
+/**
+ * @file util_ldap.h
+ * @brief Apache LDAP library
+ */
+
 #ifndef UTIL_LDAP_H
 #define UTIL_LDAP_H
 
-#include <apr_ldap.h>
+/* APR header files */
+#include "apr.h"
+#include "apr_thread_mutex.h"
+#include "apr_thread_rwlock.h"
+#include "apr_tables.h"
+#include "apr_time.h"
+#include "apr_ldap.h"
+
+#if APR_HAS_MICROSOFT_LDAPSDK
+#define AP_LDAP_IS_SERVER_DOWN(s)                ((s) == LDAP_SERVER_DOWN \
+                ||(s) == LDAP_UNAVAILABLE)
+#else
+#define AP_LDAP_IS_SERVER_DOWN(s)                ((s) == LDAP_SERVER_DOWN)
+#endif
+
+#if APR_HAS_SHARED_MEMORY
+#include "apr_rmm.h"
+#include "apr_shm.h"
+#endif
 
 /* this whole thing disappears if LDAP is not enabled */
-#ifdef APU_HAS_LDAP
-
-/* APR header files */
-#include <apr_thread_mutex.h>
-#include <apr_thread_rwlock.h>
-#include <apr_tables.h>
-#include <apr_time.h>
+#if APR_HAS_LDAP
 
 /* Apache header files */
 #include "ap_config.h"
@@ -36,14 +53,10 @@
 #include "http_log.h"
 #include "http_protocol.h"
 #include "http_request.h"
+#include "apr_optional.h"
 
-#if APR_HAS_SHARED_MEMORY
-#include "apr_rmm.h"
-#include "apr_shm.h"
-#endif
-
-/* Create a set of LDAP_DECLARE(type), LDLDAP_DECLARE(type) and 
- * LDAP_DECLARE_DATA with appropriate export and import tags for the platform
+/* Create a set of LDAP_DECLARE macros with appropriate export 
+ * and import tags for the platform
  */
 #if !defined(WIN32)
 #define LDAP_DECLARE(type)            type
@@ -63,6 +76,10 @@
 #define LDAP_DECLARE_DATA             __declspec(dllimport)
 #endif
 
+#ifdef __cplusplus
+extern "C" {
+#endif
+
 /*
  * LDAP Connections
  */
@@ -91,7 +108,8 @@ typedef struct util_ldap_connection_t {
     const char *binddn;                 /* DN to bind to server (can be NULL) */
     const char *bindpw;                 /* Password to bind to server (can be NULL) */
 
-    int secure;                         /* True if use SSL connection */
+    int secure;                         /* SSL/TLS mode of the connection */
+    apr_array_header_t *client_certs;   /* Client certificates on this connection */
 
     const char *reason;                 /* Reason for an error failure */
 
@@ -114,9 +132,11 @@ typedef struct util_ldap_state_t {
     long compare_cache_size;    /* Size (in entries) of compare cache */
 
     struct util_ldap_connection_t *connections;
-    char *cert_auth_file; 
-    int   cert_file_type;
-    int   ssl_support;
+    int   ssl_supported;
+    apr_array_header_t *global_certs;  /* Global CA certificates */
+    apr_array_header_t *client_certs;  /* Client certificates */
+    int   secure;
+    int   secure_set;
 
 #if APR_HAS_SHARED_MEMORY
     apr_shm_t *cache_shm;
@@ -126,6 +146,8 @@ typedef struct util_ldap_state_t {
     /* cache ald */
     void *util_ldap_cache;
     char *lock_file;           /* filename for shm lock mutex */
+    long  connectionTimeout;
+    int   verify_svr_cert;
 
 } util_ldap_state_t;
 
@@ -142,8 +164,8 @@ typedef struct util_ldap_state_t {
  * @deffunc int util_ldap_connection_open(request_rec *r,
  *                                        util_ldap_connection_t *ldc)
  */
-LDAP_DECLARE(int) util_ldap_connection_open(request_rec *r, 
-                                            util_ldap_connection_t *ldc);
+APR_DECLARE_OPTIONAL_FN(int,uldap_connection_open,(request_rec *r, 
+                                            util_ldap_connection_t *ldc));
 
 /**
  * Close a connection to an LDAP server
@@ -154,7 +176,7 @@ LDAP_DECLARE(int) util_ldap_connection_open(request_rec *r,
  *      structure, using apr_ldap_open_connection().
  * @deffunc util_ldap_close_connection(util_ldap_connection_t *ldc)
  */
-LDAP_DECLARE(void) util_ldap_connection_close(util_ldap_connection_t *ldc);
+APR_DECLARE_OPTIONAL_FN(void,uldap_connection_close,(util_ldap_connection_t *ldc));
 
 /**
  * Unbind a connection to an LDAP server
@@ -165,7 +187,7 @@ LDAP_DECLARE(void) util_ldap_connection_close(util_ldap_connection_t *ldc);
  *      connection back to a known state.
  * @deffunc apr_status_t util_ldap_connection_unbind(util_ldap_connection_t *ldc)
  */
-LDAP_DECLARE_NONSTD(apr_status_t) util_ldap_connection_unbind(void *param);
+APR_DECLARE_OPTIONAL_FN(apr_status_t,uldap_connection_unbind,(void *param));
 
 /**
  * Cleanup a connection to an LDAP server
@@ -175,7 +197,7 @@ LDAP_DECLARE_NONSTD(apr_status_t) util_ldap_connection_unbind(void *param);
  *      LDAP connections when the server is finished with them.
  * @deffunc apr_status_t util_ldap_connection_cleanup(util_ldap_connection_t *ldc)
  */
-LDAP_DECLARE_NONSTD(apr_status_t) util_ldap_connection_cleanup(void *param);
+APR_DECLARE_OPTIONAL_FN(apr_status_t,uldap_connection_cleanup,(void *param));
 
 /**
  * Find a connection in a list of connections
@@ -194,10 +216,9 @@ LDAP_DECLARE_NONSTD(apr_status_t) util_ldap_connection_cleanup(void *param);
  *                                                           const char *binddn, const char *bindpw, deref_options deref,
  *                                                           int netscapessl, int starttls)
  */
-LDAP_DECLARE(util_ldap_connection_t *) util_ldap_connection_find(request_rec *r, const char *host, int port,
+APR_DECLARE_OPTIONAL_FN(util_ldap_connection_t *,uldap_connection_find,(request_rec *r, const char *host, int port,
                                                   const char *binddn, const char *bindpw, deref_options deref,
-                                                  int secure);
-
+                                                  int secure));
 
 /**
  * Compare two DNs for sameness
@@ -217,9 +238,9 @@ LDAP_DECLARE(util_ldap_connection_t *) util_ldap_connection_find(request_rec *r,
  *                                        const char *url, const char *dn, const char *reqdn,
  *                                        int compare_dn_on_server)
  */
-LDAP_DECLARE(int) util_ldap_cache_comparedn(request_rec *r, util_ldap_connection_t *ldc, 
+APR_DECLARE_OPTIONAL_FN(int,uldap_cache_comparedn,(request_rec *r, util_ldap_connection_t *ldc, 
                               const char *url, const char *dn, const char *reqdn, 
-                              int compare_dn_on_server);
+                              int compare_dn_on_server));
 
 /**
  * A generic LDAP compare function
@@ -234,8 +255,8 @@ LDAP_DECLARE(int) util_ldap_cache_comparedn(request_rec *r, util_ldap_connection
  * @deffunc int util_ldap_cache_compare(request_rec *r, util_ldap_connection_t *ldc,
  *                                      const char *url, const char *dn, const char *attrib, const char *value)
  */
-LDAP_DECLARE(int) util_ldap_cache_compare(request_rec *r, util_ldap_connection_t *ldc,
-                            const char *url, const char *dn, const char *attrib, const char *value);
+APR_DECLARE_OPTIONAL_FN(int,uldap_cache_compare,(request_rec *r, util_ldap_connection_t *ldc,
+                            const char *url, const char *dn, const char *attrib, const char *value));
 
 /**
  * Checks a username/password combination by binding to the LDAP server
@@ -256,12 +277,12 @@ LDAP_DECLARE(int) util_ldap_cache_compare(request_rec *r, util_ldap_connection_t
  *                                          char *url, const char *basedn, int scope, char **attrs,
  *                                          char *filter, char *bindpw, char **binddn, char ***retvals)
  */
-LDAP_DECLARE(int) util_ldap_cache_checkuserid(request_rec *r, util_ldap_connection_t *ldc,
+APR_DECLARE_OPTIONAL_FN(int,uldap_cache_checkuserid,(request_rec *r, util_ldap_connection_t *ldc,
                               const char *url, const char *basedn, int scope, char **attrs,
-                              const char *filter, const char *bindpw, const char **binddn, const char ***retvals);
+                              const char *filter, const char *bindpw, const char **binddn, const char ***retvals));
 
 /**
- * Retrieves the LDAP DN of the user without the need to know user password
+ * Searches for a specified user object in an LDAP directory
  * @param r The request record
  * @param ldc The LDAP connection being used.
  * @param url The URL of the LDAP connection - used for deciding which cache to use.
@@ -272,20 +293,21 @@ LDAP_DECLARE(int) util_ldap_cache_checkuserid(request_rec *r, util_ldap_connecti
  *               exactly one user for the check to be successful.
  * @param binddn The DN of the user will be returned in this variable.
  * @param retvals The values corresponding to the attributes requested in the attrs array.
- * @tip The filter supplied will be searched for. A single entry matching the search is returned.
+ * @tip The filter supplied will be searched for. If a single entry is returned, an attempt
+ *      is made to bind as that user. If this bind succeeds, the user is not validated.
  * @deffunc int util_ldap_cache_getuserdn(request_rec *r, util_ldap_connection_t *ldc,
  *                                          char *url, const char *basedn, int scope, char **attrs,
  *                                          char *filter, char **binddn, char ***retvals)
  */
-LDAP_DECLARE(int) util_ldap_cache_getuserdn(request_rec *r, util_ldap_connection_t *ldc,
+APR_DECLARE_OPTIONAL_FN(int,uldap_cache_getuserdn,(request_rec *r, util_ldap_connection_t *ldc,
                               const char *url, const char *basedn, int scope, char **attrs,
-                              const char *filter, const char **binddn, const char ***retvals);
+                              const char *filter, const char **binddn, const char ***retvals));
 
 /**
  * Checks if SSL support is available in mod_ldap
  * @deffunc int util_ldap_ssl_supported(request_rec *r)
  */
-LDAP_DECLARE(int) util_ldap_ssl_supported(request_rec *r);
+APR_DECLARE_OPTIONAL_FN(int,uldap_ssl_supported,(request_rec *r));
 
 /* from apr_ldap_cache.c */
 
@@ -312,6 +334,8 @@ apr_status_t util_ldap_cache_init(apr_pool_t *pool, util_ldap_state_t *st);
  * @deffunc char *util_ald_cache_display(apr_pool_t *pool, util_ldap_state_t *st)
  */
 char *util_ald_cache_display(request_rec *r, util_ldap_state_t *st);
-
-#endif /* APU_HAS_LDAP */
+#ifdef __cplusplus
+}
+#endif
+#endif /* APR_HAS_LDAP */
 #endif /* UTIL_LDAP_H */
