@@ -1,9 +1,9 @@
-/* Copyright 1999-2005 The Apache Software Foundation or its licensors, as
- * applicable.
- *
- * Licensed under the Apache License, Version 2.0 (the "License");
- * you may not use this file except in compliance with the License.
- * You may obtain a copy of the License at
+/* Licensed to the Apache Software Foundation (ASF) under one or more
+ * contributor license agreements.  See the NOTICE file distributed with
+ * this work for additional information regarding copyright ownership.
+ * The ASF licenses this file to You under the Apache License, Version 2.0
+ * (the "License"); you may not use this file except in compliance with
+ * the License.  You may obtain a copy of the License at
  *
  *     http://www.apache.org/licenses/LICENSE-2.0
  *
@@ -70,6 +70,7 @@ module AP_MODULE_DECLARE_DATA autoindex_module;
 #define IGNORE_CLIENT       (1 << 15)
 #define IGNORE_CASE         (1 << 16)
 #define EMIT_XHTML          (1 << 17)
+#define SHOW_FORBIDDEN      (1 << 18)
 
 #define K_NOADJUST 0
 #define K_ADJUST 1
@@ -117,6 +118,8 @@ typedef struct ai_desc_t {
 typedef struct autoindex_config_struct {
 
     char *default_icon;
+    char *style_sheet;
+    char *head_insert;
     apr_int32_t opts;
     apr_int32_t incremented_opts;
     apr_int32_t decremented_opts;
@@ -136,6 +139,8 @@ typedef struct autoindex_config_struct {
     apr_array_header_t *hdr_list;
     apr_array_header_t *rdme_list;
 
+    char *ctype;
+    char *charset;
 } autoindex_config_rec;
 
 static char c_by_encoding, c_by_type, c_by_path;
@@ -151,9 +156,31 @@ static char c_by_encoding, c_by_type, c_by_path;
  */
 static void emit_preamble(request_rec *r, int xhtml, const char *title)
 {
-    ap_rvputs(r, xhtml ? DOCTYPE_XHTML_1_0T : DOCTYPE_HTML_3_2,
-              "<html>\n <head>\n  <title>Index of ", title,
-              "</title>\n </head>\n <body>\n", NULL);
+    autoindex_config_rec *d;
+
+    d = (autoindex_config_rec *) ap_get_module_config(r->per_dir_config,
+                                                      &autoindex_module);
+
+    if (xhtml) {
+        ap_rvputs(r, DOCTYPE_XHTML_1_0T,
+                  "<html xmlns=\"http://www.w3.org/1999/xhtml\">\n"
+                  " <head>\n  <title>Index of ", title,
+                  "</title>\n", NULL);
+    } else {
+        ap_rvputs(r, DOCTYPE_HTML_3_2,
+                  "<html>\n <head>\n"
+                  "  <title>Index of ", title,
+                  "</title>\n", NULL);
+    }
+
+    if (d->style_sheet != NULL) {
+        ap_rvputs(r, "  <link rel=\"stylesheet\" href=\"", d->style_sheet,
+                "\" type=\"text/css\"", xhtml ? " />\n" : ">\n", NULL);
+    }
+    if (d->head_insert != NULL) {
+        ap_rputs(d->head_insert, r);
+    }
+    ap_rvputs(r, " </head>\n <body>\n", NULL);
 }
 
 static void push_item(apr_array_header_t *arr, char *type, const char *to,
@@ -300,8 +327,9 @@ static const char *add_readme(cmd_parms *cmd, void *d, const char *name)
     return NULL;
 }
 
-static const char *add_opts(cmd_parms *cmd, void *d, const char *optstr)
+static const char *add_opts(cmd_parms *cmd, void *d, int argc, char *const argv[])
 {
+    int i;
     char *w;
     apr_int32_t opts;
     apr_int32_t opts_add;
@@ -312,10 +340,11 @@ static const char *add_opts(cmd_parms *cmd, void *d, const char *optstr)
     opts = d_cfg->opts;
     opts_add = d_cfg->incremented_opts;
     opts_remove = d_cfg->decremented_opts;
-    while (optstr[0]) {
-        int option = 0;
 
-        w = ap_getword_conf(cmd->pool, &optstr);
+    for (i = 0; i < argc; i++) {
+        int option = 0;
+        w = argv[i];
+
         if ((*w == '+') || (*w == '-')) {
             action = *(w++);
         }
@@ -373,6 +402,9 @@ static const char *add_opts(cmd_parms *cmd, void *d, const char *optstr)
         else if (!strcasecmp(w, "XHTML")) {
             option = EMIT_XHTML;
         }
+        else if (!strcasecmp(w, "ShowForbidden")) {
+            option = SHOW_FORBIDDEN;
+        }
         else if (!strcasecmp(w, "None")) {
             if (action != '\0') {
                 return "Cannot combine '+' or '-' with 'None' keyword";
@@ -459,6 +491,12 @@ static const char *add_opts(cmd_parms *cmd, void *d, const char *optstr)
                 d_cfg->desc_adjust = K_NOADJUST;
             }
         }
+        else if (!strncasecmp(w, "Type=", 5)) {
+            d_cfg->ctype = apr_pstrdup(cmd->pool, &w[5]);
+        }
+        else if (!strncasecmp(w, "Charset=", 8)) {
+            d_cfg->charset = apr_pstrdup(cmd->pool, &w[8]);
+        }
         else {
             return "Invalid directory indexing option";
         }
@@ -539,8 +577,8 @@ static const command_rec autoindex_cmds[] =
     AP_INIT_ITERATE2("AddAltByEncoding", add_alt, BY_ENCODING, DIR_CMD_PERMS,
                      "alternate descriptive text followed by one or more "
                      "content encodings"),
-    AP_INIT_RAW_ARGS("IndexOptions", add_opts, NULL, DIR_CMD_PERMS,
-                     "one or more index options [+|-][]"),
+    AP_INIT_TAKE_ARGV("IndexOptions", add_opts, NULL, DIR_CMD_PERMS,
+                      "one or more index options [+|-][]"),
     AP_INIT_TAKE2("IndexOrderDefault", set_default_order, NULL, DIR_CMD_PERMS,
                   "{Ascending,Descending} {Name,Size,Description,Date}"),
     AP_INIT_ITERATE("IndexIgnore", add_ignore, NULL, DIR_CMD_PERMS,
@@ -557,6 +595,12 @@ static const command_rec autoindex_cmds[] =
     AP_INIT_TAKE1("DefaultIcon", ap_set_string_slot,
                   (void *)APR_OFFSETOF(autoindex_config_rec, default_icon),
                   DIR_CMD_PERMS, "an icon URL"),
+    AP_INIT_TAKE1("IndexStyleSheet", ap_set_string_slot,
+                  (void *)APR_OFFSETOF(autoindex_config_rec, style_sheet),
+                  DIR_CMD_PERMS, "URL to style sheet"),
+    AP_INIT_TAKE1("IndexHeadInsert", ap_set_string_slot,
+                  (void *)APR_OFFSETOF(autoindex_config_rec, head_insert),
+                  DIR_CMD_PERMS, "String to insert in HTML HEAD section"),
     {NULL}
 };
 
@@ -595,9 +639,16 @@ static void *merge_autoindex_configs(apr_pool_t *p, void *basev, void *addv)
     new = (autoindex_config_rec *) apr_pcalloc(p, sizeof(autoindex_config_rec));
     new->default_icon = add->default_icon ? add->default_icon
                                           : base->default_icon;
+    new->style_sheet = add->style_sheet ? add->style_sheet
+                                          : base->style_sheet;
+    new->head_insert = add->head_insert ? add->head_insert
+                                          : base->head_insert;
     new->icon_height = add->icon_height ? add->icon_height : base->icon_height;
     new->icon_width = add->icon_width ? add->icon_width : base->icon_width;
 
+    new->ctype = add->ctype ? add->ctype : base->ctype;
+    new->charset = add->charset ? add->charset : base->charset;
+
     new->alt_list = apr_array_append(p, add->alt_list, base->alt_list);
     new->ign_list = apr_array_append(p, add->ign_list, base->ign_list);
     new->hdr_list = apr_array_append(p, add->hdr_list, base->hdr_list);
@@ -774,7 +825,7 @@ static char *find_default_item(char *bogus_name, apr_array_header_t *list)
  */
 
 #ifdef CASE_BLIND_FILESYSTEM
-#define MATCH_FLAGS FNM_CASE_BLIND
+#define MATCH_FLAGS APR_FNM_CASE_BLIND
 #else
 #define MATCH_FLAGS 0
 #endif
@@ -1252,28 +1303,27 @@ static struct ent *make_autoindex_entry(const apr_finfo_t *dirent,
 {
     request_rec *rr;
     struct ent *p;
+    int show_forbidden = 0;
 
     /* Dot is ignored, Parent is handled by make_parent_entry() */
     if ((dirent->name[0] == '.') && (!dirent->name[1]
         || ((dirent->name[1] == '.') && !dirent->name[2])))
         return (NULL);
 
-#ifndef CASE_BLIND_FILESYSTEM
+    /*
+     * On some platforms, the match must be case-blind.  This is really
+     * a factor of the filesystem involved, but we can't detect that
+     * reliably - so we have to granularise at the OS level.
+     */
     if (pattern && (apr_fnmatch(pattern, dirent->name,
-                                FNM_NOESCAPE | FNM_PERIOD)
-                        != APR_SUCCESS))
+                                APR_FNM_NOESCAPE | APR_FNM_PERIOD
+#ifdef CASE_BLIND_FILESYSTEM
+                                | APR_FNM_CASE_BLIND
+#endif
+                                )
+                    != APR_SUCCESS)) {
         return (NULL);
-#else  /* !CASE_BLIND_FILESYSTEM */
-        /*
-         * On some platforms, the match must be case-blind.  This is really
-         * a factor of the filesystem involved, but we can't detect that
-         * reliably - so we have to granularise at the OS level.
-         */
-    if (pattern && (apr_fnmatch(pattern, dirent->name,
-                                FNM_NOESCAPE | FNM_PERIOD | FNM_CASE_BLIND)
-                        != APR_SUCCESS))
-        return (NULL);
-#endif /* !CASE_BLIND_FILESYSTEM */
+    }
 
     if (ignore_entry(d, ap_make_full_path(r->pool,
                                           r->filename, dirent->name))) {
@@ -1284,9 +1334,15 @@ static struct ent *make_autoindex_entry(const apr_finfo_t *dirent,
         return (NULL);
     }
 
+    if((autoindex_opts & SHOW_FORBIDDEN)
+        && (rr->status == HTTP_UNAUTHORIZED || rr->status == HTTP_FORBIDDEN)) {
+        show_forbidden = 1;
+    }
+
     if ((rr->finfo.filetype != APR_DIR && rr->finfo.filetype != APR_REG)
         || !(rr->status == OK || ap_is_HTTP_SUCCESS(rr->status)
-                              || ap_is_HTTP_REDIRECT(rr->status))) {
+                              || ap_is_HTTP_REDIRECT(rr->status)
+                              || show_forbidden == 1)) {
         ap_destroy_sub_req(rr);
         return (NULL);
     }
@@ -1706,9 +1762,9 @@ static void output_directories(struct ent **ar, int n,
                                                         desc_width), NULL);
                     }
                 }
-            }
-            else {
-                ap_rputs("</td><td>&nbsp;", r);
+                else {
+                    ap_rputs("</td><td>&nbsp;", r);
+                }
             }
             ap_rputs("</td></tr>\n", r);
         }
@@ -1792,8 +1848,9 @@ static void output_directories(struct ent **ar, int n,
             ap_rputc('\n', r);
         }
         else {
-            ap_rvputs(r, "<li><a href=\"", anchor, "\"> ", t2,
-                         "</a></li>\n", NULL);
+            ap_rvputs(r, "<li><a href=\"", anchor, "\"> ",
+                      ap_escape_html(scratch, t2),
+                      "</a></li>\n", NULL);
         }
     }
     if (autoindex_opts & TABLE_INDEXING) {
@@ -1914,9 +1971,11 @@ static int dsortf(struct ent **e1, struct ent **e2)
     if (!result && c1->ignore_case) {
         result = strcasecmp (c1->name, c2->name);
     }
+
     if (!result) {
         result = strcmp (c1->name, c2->name);
     }
+
     return result;
 }
 
@@ -1941,6 +2000,8 @@ static int index_directory(request_rec *r,
     char *colargs;
     char *fullpath;
     apr_size_t dirpathlen;
+    char *ctype = "text/html";
+    char *charset;
 
     if ((status = apr_dir_open(&thedir, name, r->pool)) != APR_SUCCESS) {
         ap_log_rerror(APLOG_MARK, APLOG_ERR, status, r,
@@ -1948,11 +2009,27 @@ static int index_directory(request_rec *r,
         return HTTP_FORBIDDEN;
     }
 
+    if (autoindex_conf->ctype) {
+        ctype = autoindex_conf->ctype;
+    }
+    if (autoindex_conf->charset) {
+        charset = autoindex_conf->charset;
+    }
+    else {
 #if APR_HAS_UNICODE_FS
-    ap_set_content_type(r, "text/html;charset=utf-8");
+        charset = "UTF-8";
 #else
-    ap_set_content_type(r, "text/html");
+        charset = "ISO-8859-1";
 #endif
+    }
+    if (*charset) {
+        ap_set_content_type(r, apr_pstrcat(r->pool, ctype, ";charset=",
+                            charset, NULL));
+    }
+    else {
+        ap_set_content_type(r, ctype);
+    }
+
     if (autoindex_opts & TRACK_MODIFIED) {
         ap_update_mtime(r, r->finfo.mtime);
         ap_set_last_modified(r);
@@ -1990,24 +2067,30 @@ static int index_directory(request_rec *r,
         qstring = r->args;
 
         while (qstring && *qstring) {
-            if (qstring[0] == 'C' && qstring[1] == '='
-                    && qstring[2] && strchr(K_VALID, qstring[2])
-                    && (qstring[3] == '&' || qstring[3] == ';'
-                        || !qstring[3])) {
+
+            /* C= First Sort key Column (N, M, S, D) */
+            if (   qstring[0] == 'C' && qstring[1] == '='
+                && qstring[2] && strchr(K_VALID, qstring[2])
+                && (   qstring[3] == '&' || qstring[3] == ';'
+                    || !qstring[3])) {
                 keyid = qstring[2];
                 qstring += qstring[3] ? 4 : 3;
             }
-            else if (qstring[0] == 'O' && qstring[1] == '='
-                     && ((qstring[2] == D_ASCENDING)
+
+            /* O= Sort order (A, D) */
+            else if (   qstring[0] == 'O' && qstring[1] == '='
+                     && (   (qstring[2] == D_ASCENDING)
                          || (qstring[2] == D_DESCENDING))
-                     && (qstring[3] == '&' || qstring[3] == ';'
+                     && (   qstring[3] == '&' || qstring[3] == ';'
                          || !qstring[3])) {
                 direction = qstring[2];
                 qstring += qstring[3] ? 4 : 3;
             }
-            else if (qstring[0] == 'F' && qstring[1] == '='
+
+            /* F= Output Format (0 plain, 1 fancy (pre), 2 table) */
+            else if (   qstring[0] == 'F' && qstring[1] == '='
                      && qstring[2] && strchr("012", qstring[2])
-                     && (qstring[3] == '&' || qstring[3] == ';'
+                     && (   qstring[3] == '&' || qstring[3] == ';'
                          || !qstring[3])) {
                 if (qstring[2] == '0') {
                     autoindex_opts &= ~(FANCY_INDEXING | TABLE_INDEXING);
@@ -2023,9 +2106,11 @@ static int index_directory(request_rec *r,
                 fval[3] = qstring[2];
                 qstring += qstring[3] ? 4 : 3;
             }
-            else if (qstring[0] == 'V' && qstring[1] == '='
+
+            /* V= Version sort (0, 1) */
+            else if (   qstring[0] == 'V' && qstring[1] == '='
                      && (qstring[2] == '0' || qstring[2] == '1')
-                     && (qstring[3] == '&' || qstring[3] == ';'
+                     && (   qstring[3] == '&' || qstring[3] == ';'
                          || !qstring[3])) {
                 if (qstring[2] == '0') {
                     autoindex_opts &= ~VERSION_SORT;
@@ -2037,6 +2122,8 @@ static int index_directory(request_rec *r,
                 vval[3] = qstring[2];
                 qstring += qstring[3] ? 4 : 3;
             }
+
+            /* P= wildcard pattern (*.foo) */
             else if (qstring[0] == 'P' && qstring[1] == '=') {
                 const char *eos = qstring += 2; /* for efficiency */
 
@@ -2067,7 +2154,9 @@ static int index_directory(request_rec *r,
                     qstring = NULL;
                 }
             }
-            else {              /* Syntax error?  Ignore the remainder! */
+
+            /* Syntax error?  Ignore the remainder! */
+            else {
                 qstring = NULL;
             }
         }
@@ -2201,7 +2290,8 @@ static int handle_autoindex(request_rec *r)
     }
     else {
         ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
-                      "Directory index forbidden by rule: %s", r->filename);
+                      "Directory index forbidden by "
+                      "Options directive: %s", r->filename);
         return HTTP_FORBIDDEN;
     }
 }
