@@ -1,9 +1,9 @@
-/* Copyright 1999-2005 The Apache Software Foundation or its licensors, as
- * applicable.
- *
- * Licensed under the Apache License, Version 2.0 (the "License");
- * you may not use this file except in compliance with the License.
- * You may obtain a copy of the License at
+/* Licensed to the Apache Software Foundation (ASF) under one or more
+ * contributor license agreements.  See the NOTICE file distributed with
+ * this work for additional information regarding copyright ownership.
+ * The ASF licenses this file to You under the Apache License, Version 2.0
+ * (the "License"); you may not use this file except in compliance with
+ * the License.  You may obtain a copy of the License at
  *
  *     http://www.apache.org/licenses/LICENSE-2.0
  *
@@ -73,9 +73,12 @@
 #endif
 
 
-/* The default directory in user's home dir */
+/*
+ * The default directory in user's home dir
+ * In the default install, the module is disabled
+ */
 #ifndef DEFAULT_USER_DIR
-#define DEFAULT_USER_DIR "public_html"
+#define DEFAULT_USER_DIR NULL
 #endif
 
 module AP_MODULE_DECLARE_DATA userdir_module;
@@ -256,7 +259,7 @@ static int translate_userdir(request_rec *r)
                  * Crummy hack. Need to figure out whether we have been
                  * redirected to a URL or to a file on some drive. Since I
                  * know of no protocols that are a single letter, ignore
-                 * a : as the first or second character, and assume a file 
+                 * a : as the first or second character, and assume a file
                  * was specified
                  */
                 if (strchr(x + 2, ':'))
@@ -283,7 +286,7 @@ static int translate_userdir(request_rec *r)
 #if APR_HAS_USER
             char *homedir;
 
-            if (apr_get_home_directory(&homedir, w, r->pool) == APR_SUCCESS) {
+            if (apr_uid_homepath_get(&homedir, w, r->pool) == APR_SUCCESS) {
                 filename = apr_pstrcat(r->pool, homedir, "/", userdir, NULL);
             }
 #else
@@ -297,7 +300,7 @@ static int translate_userdir(request_rec *r)
          * anyway, in the hope that some handler might handle it. This can be
          * used, for example, to run a CGI script for the user.
          */
-        if (filename && (!*userdirs 
+        if (filename && (!*userdirs
                       || ((rv = apr_stat(&statbuf, filename, APR_FINFO_MIN,
                                          r->pool)) == APR_SUCCESS
                                              || rv == APR_INCOMPLETE))) {
@@ -334,12 +337,12 @@ static ap_unix_identity_t *get_suexec_id_doer(const request_rec *r)
         return NULL;
     }
 
-    if (apr_get_userid(&ugid->uid, &ugid->gid, username, r->pool) != APR_SUCCESS) {
+    if (apr_uid_get(&ugid->uid, &ugid->gid, username, r->pool) != APR_SUCCESS) {
         return NULL;
     }
 
     ugid->userdir = 1;
-#endif 
+#endif
     return ugid;
 }
 #endif /* HAVE_UNIX_SUEXEC */
