@@ -36,7 +36,13 @@
 #include "apr_pools.h"
 #include "apr_dbd_internal.h"
 
+#ifdef HAVE_FREETDS_SYBDB_H
+#include <freetds/sybdb.h>
+#endif
+#ifdef HAVE_SYBDB_H
 #include <sybdb.h>
+#endif
+
 #include <stdio.h>
 #include <sys/types.h>
 #include <regex.h>
@@ -124,7 +130,6 @@ static int dbd_freetds_select(apr_pool_t *pool, apr_dbd_t *sql,
                               const char *query, int seek)
 {
     apr_dbd_results_t *res;
-    int i;
     if (sql->trans && (sql->trans->errnum != SUCCEED)) {
         return 1;
     }
@@ -237,9 +242,10 @@ static const char *dbd_statement(apr_pool_t *pool,
 static int dbd_freetds_pselect(apr_pool_t *pool, apr_dbd_t *sql,
                                apr_dbd_results_t **results,
                                apr_dbd_prepared_t *statement,
-                               int seek, int nargs, const char **values)
+                               int seek, const char **values)
 {
-    const char *query = dbd_statement(pool, statement, nargs, values);
+    const char *query = dbd_statement(pool, statement,
+                                      statement->nargs, values);
     return dbd_freetds_select(pool, sql, results, query, seek);
 }
 static int dbd_freetds_pvselect(apr_pool_t *pool, apr_dbd_t *sql,
@@ -260,15 +266,15 @@ static int dbd_freetds_pvselect(apr_pool_t *pool, apr_dbd_t *sql,
         values[i] = va_arg(args, const char*);
     }
 
-    return dbd_freetds_pselect(pool, sql, results, statement, seek,
-                               statement->nargs, values);
+    return dbd_freetds_pselect(pool, sql, results, statement, seek, values);
 }
 static int dbd_freetds_query(apr_dbd_t *sql, int *nrows, const char *query);
 static int dbd_freetds_pquery(apr_pool_t *pool, apr_dbd_t *sql,
                               int *nrows, apr_dbd_prepared_t *statement,
-                              int nargs, const char **values)
+                              const char **values)
 {
-    const char *query = dbd_statement(pool, statement, nargs, values);
+    const char *query = dbd_statement(pool, statement,
+                                      statement->nargs, values);
     return dbd_freetds_query(sql, nrows, query);
 }
 static int dbd_freetds_pvquery(apr_pool_t *pool, apr_dbd_t *sql, int *nrows,
@@ -286,8 +292,7 @@ static int dbd_freetds_pvquery(apr_pool_t *pool, apr_dbd_t *sql, int *nrows,
     for (i = 0; i < statement->nargs; i++) {
         values[i] = va_arg(args, const char*);
     }
-    return dbd_freetds_pquery(pool, sql, nrows, statement,
-                              statement->nargs, values);
+    return dbd_freetds_pquery(pool, sql, nrows, statement, values);
 }
 
 static int dbd_freetds_get_row(apr_pool_t *pool, apr_dbd_results_t *res,
@@ -344,10 +349,10 @@ static const char *dbd_freetds_get_entry(const apr_dbd_row_t *row, int n)
     int t = dbcoltype(proc, n+1);
     int l = dbcollen(proc, n+1);
     if (dbwillconvert(t, SYBCHAR)) {
-      dbconvert(proc, t, ptr, l, SYBCHAR, row->buf, -1);
+      dbconvert(proc, t, ptr, l, SYBCHAR, (BYTE *)row->buf, -1);
       return (const char*)row->buf;
     }
-    return ptr;
+    return (char*)ptr;
 }
 
 static const char *dbd_freetds_error(apr_dbd_t *sql, int n)
@@ -442,6 +447,7 @@ static int recurse_args(apr_pool_t *pool, int n, const char *query,
 
 static int dbd_freetds_prepare(apr_pool_t *pool, apr_dbd_t *sql,
                              const char *query, const char *label,
+                             int nargs, int nvals, apr_dbd_type_e *types,
                              apr_dbd_prepared_t **statement)
 {
     apr_dbd_prepared_t *stmt;
@@ -556,14 +562,14 @@ static DBPROCESS *freetds_open(apr_pool_t *pool, const char *params,
             ++ptr;
             continue;
         }
-        for (key = ptr-1; isspace(*key); --key);
+        for (key = ptr-1; apr_isspace(*key); --key);
         klen = 0;
-        while (isalpha(*key)) {
+        while (apr_isalpha(*key)) {
             --key;
             ++klen;
         }
         ++key;
-        for (value = ptr+1; isspace(*value); ++value);
+        for (value = ptr+1; apr_isspace(*value); ++value);
 
         vlen = strcspn(value, delims);
         buf = apr_pstrndup(pool, value, vlen);        /* NULL-terminated copy */
@@ -600,9 +606,7 @@ static DBPROCESS *freetds_open(apr_pool_t *pool, const char *params,
 
     process = dbopen(login, server);
 
-    fprintf(stderr, "databaseName [%s]\n", databaseName);
-
-    if (databaseName != NULL)
+    if (process != NULL && databaseName != NULL)
     {
         dbuse(process, databaseName);
     }
@@ -642,7 +646,7 @@ static apr_status_t dbd_freetds_check_conn(apr_pool_t *pool,
     if (dbdead(handle->proc)) {
         /* try again */
         dbclose(handle->proc);
-        handle->proc = freetds_open(handle->pool, handle->params);
+        handle->proc = freetds_open(handle->pool, handle->params, NULL);
         if (!handle->proc || dbdead(handle->proc)) {
             return APR_EGENERAL;
         }
@@ -685,6 +689,11 @@ static apr_status_t freetds_term(void *dummy)
     regfree(&dbd_freetds_find_arg);
     return APR_SUCCESS;
 }
+static int freetds_err_handler(DBPROCESS *dbproc, int severity, int dberr,
+                               int oserr, char *dberrstr, char *oserrstr)
+{
+    return INT_CANCEL; /* never exit */
+}
 static void dbd_freetds_init(apr_pool_t *pool)
 {
     int rv = regcomp(&dbd_freetds_find_arg,
@@ -695,6 +704,7 @@ static void dbd_freetds_init(apr_pool_t *pool)
         fprintf(stderr, "regcomp failed: %s\n", errmsg);
     }
     dbinit();
+    dberrhandle(freetds_err_handler);
     apr_pool_cleanup_register(pool, NULL, freetds_term, apr_pool_cleanup_null);
 }
 
