@@ -92,7 +92,7 @@ static int dbd_pgsql_select(apr_pool_t *pool, apr_dbd_t *sql,
         if (TXN_IGNORE_ERRORS(sql->trans)) {
             PGresult *res = PQexec(sql->conn, "SAVEPOINT APR_DBD_TXN_SP");
             if (res) {
-                ret = PQresultStatus(res);
+                int ret = PQresultStatus(res);
                 PQclear(res);
                 if (!dbd_pgsql_is_success(ret)) {
                     sql->trans->errnum = ret;
@@ -118,7 +118,7 @@ static int dbd_pgsql_select(apr_pool_t *pool, apr_dbd_t *sql,
                 PGresult *res = PQexec(sql->conn,
                                        "ROLLBACK TO SAVEPOINT APR_DBD_TXN_SP");
                 if (res) {
-                    ret = PQresultStatus(res);
+                    int ret = PQresultStatus(res);
                     PQclear(res);
                     if (!dbd_pgsql_is_success(ret)) {
                         sql->trans->errnum = ret;
@@ -136,7 +136,7 @@ static int dbd_pgsql_select(apr_pool_t *pool, apr_dbd_t *sql,
                 PGresult *res = PQexec(sql->conn,
                                        "RELEASE SAVEPOINT APR_DBD_TXN_SP");
                 if (res) {
-                    ret = PQresultStatus(res);
+                    int ret = PQresultStatus(res);
                     PQclear(res);
                     if (!dbd_pgsql_is_success(ret)) {
                         sql->trans->errnum = ret;
@@ -162,7 +162,7 @@ static int dbd_pgsql_select(apr_pool_t *pool, apr_dbd_t *sql,
         if (TXN_IGNORE_ERRORS(sql->trans)) {
             PGresult *res = PQexec(sql->conn, "SAVEPOINT APR_DBD_TXN_SP");
             if (res) {
-                ret = PQresultStatus(res);
+                int ret = PQresultStatus(res);
                 PQclear(res);
                 if (!dbd_pgsql_is_success(ret)) {
                     sql->trans->errnum = ret;
@@ -177,7 +177,7 @@ static int dbd_pgsql_select(apr_pool_t *pool, apr_dbd_t *sql,
                 PGresult *res = PQexec(sql->conn,
                                        "ROLLBACK TO SAVEPOINT APR_DBD_TXN_SP");
                 if (res) {
-                    ret = PQresultStatus(res);
+                    int ret = PQresultStatus(res);
                     PQclear(res);
                     if (!dbd_pgsql_is_success(ret)) {
                         sql->trans->errnum = ret;
@@ -195,7 +195,7 @@ static int dbd_pgsql_select(apr_pool_t *pool, apr_dbd_t *sql,
                 PGresult *res = PQexec(sql->conn,
                                        "RELEASE SAVEPOINT APR_DBD_TXN_SP");
                 if (res) {
-                    ret = PQresultStatus(res);
+                    int ret = PQresultStatus(res);
                     PQclear(res);
                     if (!dbd_pgsql_is_success(ret)) {
                         sql->trans->errnum = ret;
@@ -218,7 +218,12 @@ static int dbd_pgsql_select(apr_pool_t *pool, apr_dbd_t *sql,
 
 static const char *dbd_pgsql_get_name(const apr_dbd_results_t *res, int n)
 {
-    return (res->res ? PQfname(res->res, n) : NULL);
+    if (res->res) {
+        if ((n>=0) && (PQnfields(res->res) > n)) {
+            return PQfname(res->res,n);
+        }
+    }
+    return NULL;
 }
 
 static int dbd_pgsql_get_row(apr_pool_t *pool, apr_dbd_results_t *res,
@@ -258,7 +263,7 @@ static int dbd_pgsql_get_row(apr_pool_t *pool, apr_dbd_results_t *res,
     }
 
     if (res->random) {
-        if (row->n >= res->ntuples) {
+        if ((row->n >= 0) && (size_t)row->n >= res->ntuples) {
             *rowp = NULL;
             apr_pool_cleanup_run(pool, res->res, clear_result);
             res->res = NULL;
@@ -266,7 +271,7 @@ static int dbd_pgsql_get_row(apr_pool_t *pool, apr_dbd_results_t *res,
         }
     }
     else {
-        if (row->n >= res->ntuples) {
+        if ((row->n >= 0) && (size_t)row->n >= res->ntuples) {
             /* no data; we have to fetch some */
             row->n -= res->ntuples;
             if (res->res != NULL) {
@@ -344,7 +349,7 @@ static apr_status_t dbd_pgsql_datum_get(const apr_dbd_row_t *row, int n,
         *(apr_uint64_t*)data = apr_atoi64(PQgetvalue(row->res->res, row->n, n));
         break;
     case APR_DBD_TYPE_FLOAT:
-        *(float*)data = atof(PQgetvalue(row->res->res, row->n, n));
+        *(float*)data = (float)atof(PQgetvalue(row->res->res, row->n, n));
         break;
     case APR_DBD_TYPE_DOUBLE:
         *(double*)data = atof(PQgetvalue(row->res->res, row->n, n));
@@ -396,7 +401,7 @@ static int dbd_pgsql_query(apr_dbd_t *sql, int *nrows, const char *query)
     if (TXN_IGNORE_ERRORS(sql->trans)) {
         PGresult *res = PQexec(sql->conn, "SAVEPOINT APR_DBD_TXN_SP");
         if (res) {
-            ret = PQresultStatus(res);
+            int ret = PQresultStatus(res);
             PQclear(res);
             if (!dbd_pgsql_is_success(ret)) {
                 sql->trans->errnum = ret;
@@ -426,7 +431,7 @@ static int dbd_pgsql_query(apr_dbd_t *sql, int *nrows, const char *query)
             PGresult *res = PQexec(sql->conn,
                                    "ROLLBACK TO SAVEPOINT APR_DBD_TXN_SP");
             if (res) {
-                ret = PQresultStatus(res);
+                int ret = PQresultStatus(res);
                 PQclear(res);
                 if (!dbd_pgsql_is_success(ret)) {
                     sql->trans->errnum = ret;
@@ -444,7 +449,7 @@ static int dbd_pgsql_query(apr_dbd_t *sql, int *nrows, const char *query)
             PGresult *res = PQexec(sql->conn,
                                    "RELEASE SAVEPOINT APR_DBD_TXN_SP");
             if (res) {
-                ret = PQresultStatus(res);
+                int ret = PQresultStatus(res);
                 PQclear(res);
                 if (!dbd_pgsql_is_success(ret)) {
                     sql->trans->errnum = ret;
@@ -477,7 +482,7 @@ static int dbd_pgsql_prepare(apr_pool_t *pool, apr_dbd_t *sql,
     char *sqlcmd;
     char *sqlptr;
     size_t length, qlen;
-    size_t i = 0;
+    int i = 0;
     const char **args;
     size_t alen;
     int ret;
@@ -611,7 +616,7 @@ static int dbd_pgsql_pquery_internal(apr_pool_t *pool, apr_dbd_t *sql,
     if (TXN_IGNORE_ERRORS(sql->trans)) {
         PGresult *res = PQexec(sql->conn, "SAVEPOINT APR_DBD_TXN_SP");
         if (res) {
-            ret = PQresultStatus(res);
+            int ret = PQresultStatus(res);
             PQclear(res);
             if (!dbd_pgsql_is_success(ret)) {
                 sql->trans->errnum = ret;
@@ -647,7 +652,7 @@ static int dbd_pgsql_pquery_internal(apr_pool_t *pool, apr_dbd_t *sql,
             PGresult *res = PQexec(sql->conn,
                                    "ROLLBACK TO SAVEPOINT APR_DBD_TXN_SP");
             if (res) {
-                ret = PQresultStatus(res);
+                int ret = PQresultStatus(res);
                 PQclear(res);
                 if (!dbd_pgsql_is_success(ret)) {
                     sql->trans->errnum = ret;
@@ -665,7 +670,7 @@ static int dbd_pgsql_pquery_internal(apr_pool_t *pool, apr_dbd_t *sql,
             PGresult *res = PQexec(sql->conn,
                                    "RELEASE SAVEPOINT APR_DBD_TXN_SP");
             if (res) {
-                ret = PQresultStatus(res);
+                int ret = PQresultStatus(res);
                 PQclear(res);
                 if (!dbd_pgsql_is_success(ret)) {
                     sql->trans->errnum = ret;
@@ -767,7 +772,7 @@ static int dbd_pgsql_pselect_internal(apr_pool_t *pool, apr_dbd_t *sql,
         if (TXN_IGNORE_ERRORS(sql->trans)) {
             PGresult *res = PQexec(sql->conn, "SAVEPOINT APR_DBD_TXN_SP");
             if (res) {
-                ret = PQresultStatus(res);
+                int ret = PQresultStatus(res);
                 PQclear(res);
                 if (!dbd_pgsql_is_success(ret)) {
                     sql->trans->errnum = ret;
@@ -803,7 +808,7 @@ static int dbd_pgsql_pselect_internal(apr_pool_t *pool, apr_dbd_t *sql,
                 PGresult *res = PQexec(sql->conn,
                                        "ROLLBACK TO SAVEPOINT APR_DBD_TXN_SP");
                 if (res) {
-                    ret = PQresultStatus(res);
+                    int ret = PQresultStatus(res);
                     PQclear(res);
                     if (!dbd_pgsql_is_success(ret)) {
                         sql->trans->errnum = ret;
@@ -822,7 +827,7 @@ static int dbd_pgsql_pselect_internal(apr_pool_t *pool, apr_dbd_t *sql,
                 PGresult *res = PQexec(sql->conn,
                                        "RELEASE SAVEPOINT APR_DBD_TXN_SP");
                 if (res) {
-                    ret = PQresultStatus(res);
+                    int ret = PQresultStatus(res);
                     PQclear(res);
                     if (!dbd_pgsql_is_success(ret)) {
                         sql->trans->errnum = ret;
@@ -849,7 +854,7 @@ static int dbd_pgsql_pselect_internal(apr_pool_t *pool, apr_dbd_t *sql,
         if (TXN_IGNORE_ERRORS(sql->trans)) {
             PGresult *res = PQexec(sql->conn, "SAVEPOINT APR_DBD_TXN_SP");
             if (res) {
-                ret = PQresultStatus(res);
+                int ret = PQresultStatus(res);
                 PQclear(res);
                 if (!dbd_pgsql_is_success(ret)) {
                     sql->trans->errnum = ret;
@@ -873,7 +878,7 @@ static int dbd_pgsql_pselect_internal(apr_pool_t *pool, apr_dbd_t *sql,
                 PGresult *res = PQexec(sql->conn,
                                        "ROLLBACK TO SAVEPOINT APR_DBD_TXN_SP");
                 if (res) {
-                    ret = PQresultStatus(res);
+                    int ret = PQresultStatus(res);
                     PQclear(res);
                     if (!dbd_pgsql_is_success(ret)) {
                         sql->trans->errnum = ret;
@@ -892,7 +897,7 @@ static int dbd_pgsql_pselect_internal(apr_pool_t *pool, apr_dbd_t *sql,
                 PGresult *res = PQexec(sql->conn,
                                        "RELEASE SAVEPOINT APR_DBD_TXN_SP");
                 if (res) {
-                    ret = PQresultStatus(res);
+                    int ret = PQresultStatus(res);
                     PQclear(res);
                     if (!dbd_pgsql_is_success(ret)) {
                         sql->trans->errnum = ret;
@@ -1191,6 +1196,16 @@ static int dbd_pgsql_transaction_mode_set(apr_dbd_transaction_t *trans,
     return trans->mode = (mode & TXN_MODE_BITS);
 }
 
+static void null_notice_receiver(void *arg, const PGresult *res)
+{
+    /* nothing */
+}
+
+static void null_notice_processor(void *arg, const char *message)
+{
+    /* nothing */
+}
+
 static apr_dbd_t *dbd_pgsql_open(apr_pool_t *pool, const char *params,
                                  const char **error)
 {
@@ -1210,6 +1225,9 @@ static apr_dbd_t *dbd_pgsql_open(apr_pool_t *pool, const char *params,
         return NULL;
     }
 
+    PQsetNoticeReceiver(conn, null_notice_receiver, NULL);
+    PQsetNoticeProcessor(conn, null_notice_processor, NULL);
+
     sql = apr_pcalloc (pool, sizeof (*sql));
 
     sql->conn = conn;
