@@ -1,9 +1,9 @@
-/* Copyright 2000-2005 The Apache Software Foundation or its licensors, as
- * applicable.
- *
- * Licensed under the Apache License, Version 2.0 (the "License");
- * you may not use this file except in compliance with the License.
- * You may obtain a copy of the License at
+/* Licensed to the Apache Software Foundation (ASF) under one or more
+ * contributor license agreements.  See the NOTICE file distributed with
+ * this work for additional information regarding copyright ownership.
+ * The ASF licenses this file to You under the Apache License, Version 2.0
+ * (the "License"); you may not use this file except in compliance with
+ * the License.  You may obtain a copy of the License at
  *
  *     http://www.apache.org/licenses/LICENSE-2.0
  *
@@ -77,7 +77,7 @@ static void apr_setup(apr_pool_t **p, apr_socket_t **sock, int *family)
     }
 
     *sock = NULL;
-    rv = apr_socket_create(sock, *family, SOCK_STREAM, *p);
+    rv = apr_socket_create(sock, *family, SOCK_STREAM, 0, *p);
     if (rv != APR_SUCCESS) {
         fprintf(stderr, "apr_socket_create()->%d/%s\n",
                 rv,
@@ -147,7 +147,7 @@ static void create_testfile(apr_pool_t *p, const char *fname)
     }
 
     rv = apr_stat(&finfo, fname, APR_FINFO_NORM, p);
-    if (rv != APR_SUCCESS && rv != APR_INCOMPLETE) {
+    if (rv != APR_SUCCESS && ! APR_STATUS_IS_INCOMPLETE(rv)) {
         fprintf(stderr, "apr_stat()->%d/%s\n",
                 rv, apr_strerror(rv, buf, sizeof buf));
         exit(1);
@@ -178,7 +178,7 @@ static int client(client_socket_mode_t socket_mode, char *host)
     struct iovec headers[3];
     struct iovec trailers[3];
     apr_size_t bytes_read;
-    apr_pollfd_t *pfd;
+    apr_pollset_t *pset;
     apr_int32_t nsocks;
     int i;
     int family;
@@ -301,11 +301,19 @@ static int client(client_socket_mode_t socket_mode, char *host)
     else {
         /* non-blocking... wooooooo */
         apr_size_t total_bytes_sent;
+        apr_pollfd_t pfd;
 
-        pfd = NULL;
-        rv = apr_poll_setup(&pfd, 1, p);
+        pset = NULL;
+        rv = apr_pollset_create(&pset, 1, p, 0);
         assert(!rv);
-        rv = apr_poll_socket_add(pfd, sock, APR_POLLOUT);
+        pfd.p = p;
+        pfd.desc_type = APR_POLL_SOCKET;
+        pfd.reqevents = APR_POLLOUT;
+        pfd.rtnevents = 0;
+        pfd.desc.s = sock;
+        pfd.client_data = NULL;
+
+        rv = apr_pollset_add(pset, &pfd);        
         assert(!rv);
 
         total_bytes_sent = 0;
@@ -336,7 +344,7 @@ static int client(client_socket_mode_t socket_mode, char *host)
                 if (APR_STATUS_IS_EAGAIN(rv)) {
                     assert(tmplen == 0);
                     nsocks = 1;
-                    tmprv = apr_poll(pfd, 1, &nsocks, -1);
+                    tmprv = apr_pollset_poll(pset, -1, &nsocks, NULL);
                     assert(!tmprv);
                     assert(nsocks == 1);
                     /* continue; */
